#![feature(round_char_boundary)]
use std::io::Read;

use ::organic::parser::document;
#[cfg(feature = "compare")]
use organic::compare_document;
#[cfg(feature = "compare")]
use organic::emacs_parse_org_document;
#[cfg(feature = "compare")]
use organic::get_emacs_version;
#[cfg(feature = "compare")]
use organic::get_org_mode_version;
#[cfg(feature = "compare")]
use organic::parser::sexp::sexp_with_padding;

#[cfg(feature = "tracing")]
use crate::init_tracing::init_telemetry;
#[cfg(feature = "tracing")]
use crate::init_tracing::shutdown_telemetry;
#[cfg(feature = "tracing")]
mod init_tracing;

#[cfg(not(feature = "tracing"))]
fn main() -> Result<(), Box<dyn std::error::Error>> {
    main_body()
}

#[cfg(feature = "tracing")]
fn main() -> Result<(), Box<dyn std::error::Error>> {
    let rt = tokio::runtime::Runtime::new()?;
    let result = rt.block_on(async {
        init_telemetry()?;
        let main_body_result = main_body();
        shutdown_telemetry()?;
        main_body_result
    });
    result
}

#[cfg_attr(feature = "tracing", tracing::instrument(ret, level = "debug"))]
fn main_body() -> Result<(), Box<dyn std::error::Error>> {
    let org_contents = read_stdin_to_string()?;
    run_compare(org_contents)
}

fn read_stdin_to_string() -> Result<String, Box<dyn std::error::Error>> {
    let mut stdin_contents = String::new();
    std::io::stdin()
        .lock()
        .read_to_string(&mut stdin_contents)?;
    Ok(stdin_contents)
}

#[cfg(feature = "compare")]
fn run_compare<P: AsRef<str>>(org_contents: P) -> Result<(), Box<dyn std::error::Error>> {
    let emacs_version = get_emacs_version()?;
    let org_mode_version = get_org_mode_version()?;
    eprintln!("Using emacs version: {}", emacs_version.trim());
    eprintln!("Using org-mode version: {}", org_mode_version.trim());
    let (remaining, rust_parsed) = document(org_contents.as_ref()).map_err(|e| e.to_string())?;
    let org_sexp = emacs_parse_org_document(org_contents.as_ref())?;
    let (_remaining, parsed_sexp) =
        sexp_with_padding(org_sexp.as_str()).map_err(|e| e.to_string())?;

    println!("{}\n\n\n", org_contents.as_ref());
    println!("{}", org_sexp);
    println!("{:#?}", rust_parsed);

    // We do the diffing after printing out both parsed forms in case the diffing panics
    let diff_result = compare_document(&parsed_sexp, &rust_parsed)?;
    diff_result.print()?;

    if diff_result.is_bad() {
        Err("Diff results do not match.")?;
    }
    if remaining != "" {
        Err(format!("There was unparsed text remaining: {}", remaining))?;
    }

    Ok(())
}

#[cfg(not(feature = "compare"))]
fn run_compare<P: AsRef<str>>(org_contents: P) -> Result<(), Box<dyn std::error::Error>> {
    eprintln!(
        "This program was built with compare disabled. Only parsing with organic, not comparing."
    );
    let (remaining, rust_parsed) = document(org_contents.as_ref()).map_err(|e| e.to_string())?;
    println!("{:#?}", rust_parsed);
    Ok(())
}
